package ru.dbotthepony.mc.otm.storage

import it.unimi.dsi.fastutil.longs.Long2ObjectAVLTreeMap
import it.unimi.dsi.fastutil.longs.Long2ObjectFunction
import it.unimi.dsi.fastutil.objects.Object2ObjectAVLTreeMap
import ru.dbotthepony.mc.otm.capability.IMatteryEnergyStorage
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import java.util.*

class RemoteTuple<T : IStorageStack>(val obj: T, val remote_id: UUID, val provider: IStorageView<T>, val local: LocalTuple<T>) {
	fun extract(amount: ImpreciseFraction, simulate: Boolean): T {
		return provider.extractStack(remote_id, amount, simulate)
	}

	fun extractCount(amount: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		return provider.extractStackCount(remote_id, amount, simulate)
	}

	override fun equals(other: Any?): Boolean {
		return other is RemoteTuple<*> && other.remote_id == remote_id || other is UUID && other == remote_id
	}

	override fun hashCode(): Int {
		return remote_id.hashCode()
	}
}

class LocalTuple<T : IStorageStack>(override val stack: T, override val id: UUID, val tuples: ArrayList<RemoteTuple<T>>) : IStorageTuple<T>

open class VirtualComponent<T : IStorageStack>(type: StorageStackType<T>) : IStorageComponent<T>, IStorageListener<T> {
	constructor(type: Class<T>) : this(StorageRegistry.get(type))

	override val storageType: StorageStackType<T> = type

	// удаленный UUID -> Кортеж
	protected val remoteByUUID = Object2ObjectAVLTreeMap<UUID, RemoteTuple<T>>()

	// локальный UUID -> Локальный кортеж
	protected val localByUUID = Object2ObjectAVLTreeMap<UUID, LocalTuple<T>>()

	// ключ -> Список Локальных кортежей
	protected val partitions = Long2ObjectAVLTreeMap<ArrayList<LocalTuple<T>>>()

	// ArrayList для скорости работы
	protected val listeners = ArrayList<IStorageListener<T>>()
	protected val consumers = ArrayList<IStorageConsumer<T>>()

	open fun add(identity: IStorage<T>) {
		if (identity is IStorageView<T>) {
			identity.addListenerAuto(this)
		} else if (identity is IStorageTrigger<T>) {
			identity.addListener(this)
		}

		if (identity is IStorageConsumer<T> && !consumers.contains(identity)) {
			consumers.add(identity)
		}
	}

	open fun remove(identity: IStorage<T>) {
		if (identity is IStorageView<T>) {
			identity.removeListenerAuto(this)
		} else if (identity is IStorageTrigger<T>) {
			identity.removeListener(this)
		}

		if (identity is IStorageConsumer<T>) {
			consumers.remove(identity)
		}
	}

	override fun addListener(listener: IStorageListener<T>): Boolean {
		if (!listeners.contains(listener)) {
			listeners.add(listener)
			return true
		}

		return false
	}

	override fun removeListener(listener: IStorageListener<T>): Boolean {
		return listeners.remove(listener)
	}

	override fun getStack(id: UUID): T {
		return localByUUID[id]?.stack ?: this.storageType.empty
	}

	override fun getStacks(): List<IStorageTuple<T>> {
		var capacity = 0
		for (list in partitions.values) capacity += list.size
		val output = ArrayList<IStorageTuple<T>>(capacity)
		for (listing in partitions.values) output.addAll(listing)
		return output
	}

	@Suppress("unchecked_cast")
	override fun addStack(stack: T, id: UUID, provider: IStorageView<T>) {
		check(!remoteByUUID.containsKey(id)) { "Already tracking tuple with id $id" }
		val items = partitions.computeIfAbsent(stack.partitionKey, Long2ObjectFunction { ArrayList() })
		var local: LocalTuple<T>? = null

		for (item in items) {
			if (item.stack.sameItem(stack)) {
				item.stack.grow(stack.count)
				local = item
				break
			}
		}

		val added = local == null

		if (local == null) {
			local = LocalTuple(stack.copy() as T, UUID.randomUUID(), ArrayList<RemoteTuple<T>>(1))
			items.add(local)
			localByUUID[local.id] = local
		}

		val remote = RemoteTuple(stack.copy() as T, id, provider, local)
		local.tuples.add(remote)
		remoteByUUID[id] = remote

		if (added) {
			for (listener in listeners) {
				listener.addStack(local.stack, local.id, this)
			}
		} else {
			for (listener in listeners) {
				listener.changeStack(local.id, local.stack.count)
			}
		}
	}

	override fun changeStack(id: UUID, newCount: ImpreciseFraction) {
		assert(newCount > ImpreciseFraction.ZERO)
		val tuple = remoteByUUID[id] ?: throw IllegalStateException("No such tuple with id $id")

		val diff = newCount - tuple.obj.count
		tuple.obj.count = newCount
		tuple.local.stack.grow(diff)

		for (listener in listeners) {
			listener.changeStack(tuple.local.id, tuple.local.stack.count)
		}
	}

	override fun removeStack(id: UUID) {
		val tuple = remoteByUUID[id] ?: throw IllegalStateException("No such tuple with id $id")
		val key = tuple.local.stack.partitionKey

		tuple.local.stack.shrink(tuple.obj.count)
		tuple.local.tuples.remove(tuple)

		remoteByUUID.remove(id)

		val a = tuple.local.stack.count <= ImpreciseFraction.ZERO
		val b = tuple.local.tuples.size == 0

		if (a || b) {
			check(a && b) { "View object is empty, but tuple list is not!" }
			localByUUID.remove(tuple.local.id)
			partitions[key]!!.remove(tuple.local)

			for (listener in listeners) {
				listener.removeStack(tuple.local.id)
			}
		}
	}

	override fun insertStack(stack: T, simulate: Boolean): T {
		var leftover = stack

		for (consumer in consumers) {
			leftover = consumer.insertStack(leftover, simulate)

			if (leftover.isEmpty) {
				return leftover
			}
		}

		return leftover
	}

	@Suppress("unchecked_cast")
	override fun extractStack(id: UUID, amount: ImpreciseFraction, simulate: Boolean): T {
		if (amount.isZero)
			return this.storageType.empty

		@Suppress("name_shadowing")
		var amount = amount
		val tuple: LocalTuple<T>? = localByUUID[id]

		if (tuple == null || amount.isZero)
			return this.storageType.empty

		if (amount <= ImpreciseFraction.MINUS_ONE)
			amount = tuple.stack.maxStackSize ?: tuple.stack.count

		val toExtract = tuple.stack.count.min(amount)
		var extracted = ImpreciseFraction.ZERO
		val copy = tuple.stack.copy() as T

		for (remote_tuple in tuple.tuples) {
			extracted += remote_tuple.extractCount(toExtract - extracted, simulate)

			if (extracted >= toExtract)
				break
		}

		if (extracted > ImpreciseFraction.ZERO) {
			copy.count = extracted
			return copy
		}

		return this.storageType.empty
	}
}

open class PoweredVirtualComponent<T : IStorageStack>(type: StorageStackType<T>, val energyProvider: () -> IMatteryEnergyStorage) : VirtualComponent<T>(type) {
	constructor(type: Class<T>, energyStorage: IMatteryEnergyStorage) : this(StorageRegistry.get(type), {energyStorage})
	constructor(type: StorageStackType<T>, energyStorage: IMatteryEnergyStorage) : this(type, {energyStorage})
	constructor(other: VirtualComponent<T>, energyStorage: IMatteryEnergyStorage) : this(other.storageType.identity, energyStorage) {
		add(other)
	}

	@Suppress("unchecked_cast")
	override fun insertStack(stack: T, simulate: Boolean): T {
		val required = storageType.energyPerOperation * stack.count
		val energy = energyProvider.invoke()
		val extracted = energy.extractEnergyInner(required, true)

		if (extracted.isZero) {
			return stack.copy() as T
		}

		if (extracted == required) {
			val leftover = super.insertStack(stack, simulate)

			if (leftover.isEmpty) {
				if (!simulate) {
					energy.extractEnergyInner(required, false)
				}

				return leftover
			}

			if (!simulate) {
				val requiredNew = storageType.energyPerOperation * stack.count - leftover.count
				energy.extractEnergyInner(requiredNew, false)
			}

			return leftover
		}

		@Suppress("name_shadowing")
		val stack = stack.copy() as T
		val oldCount = stack.count
		stack.count = extracted / storageType.energyPerOperation
		val diff = oldCount - stack.count
		val newRequired = storageType.energyPerOperation * stack.count
		val newExtracted = energy.extractEnergyInner(newRequired, true)

		if (newExtracted == newRequired) {
			val leftover = super.insertStack(stack, simulate)

			if (leftover.isEmpty) {
				if (!simulate) {
					energy.extractEnergyInner(newRequired, false)
				}

				leftover.count = diff
				return leftover
			}

			if (!simulate) {
				val requiredNew = (stack.count - leftover.count) * storageType.energyPerOperation
				energy.extractEnergyInner(requiredNew, false)
			}

			leftover.count += diff
			return leftover
		}

		return stack
	}

	override fun extractStack(id: UUID, amount: ImpreciseFraction, simulate: Boolean): T {
		val required = storageType.energyPerOperation * amount
		val energy = energyProvider.invoke()
		val extracted = energy.extractEnergyInner(required, true)

		if (extracted.isZero) {
			return storageType.empty
		}

		if (extracted == required) {
			val extractedStack = super.extractStack(id, amount, simulate)

			if (extractedStack.isEmpty) {
				return extractedStack
			}

			if (!simulate) {
				if (extractedStack.count == amount) {
					energy.extractEnergyInner(required, false)
				} else {
					energy.extractEnergyInner(extractedStack.count * storageType.energyPerOperation, false)
				}
			}

			return extractedStack
		}

		@Suppress("name_shadowing")
		val amount = required / storageType.energyPerOperation
		val extractedStack = super.extractStack(id, amount, simulate)

		if (extractedStack.isEmpty) {
			return extractedStack
		}

		if (!simulate) {
			energy.extractEnergyInner(extractedStack.count * storageType.energyPerOperation, false)
		}

		return extractedStack
	}
}
